__all__ = [
    "alias",
    "converter",
    "field_input_converter",
    "field_output_converter",
    "inout_model",
    "input_converter",
    "output_converter",
    "from_data",
    "from_stringified",
    "to_data",
    "properties",
    "fields_set",
    "get_fields",
    "mark_set_fields",
    "unmark_set_fields",
    "with_fields_set",
    "Ignored",
    "JSONSchema",
    "build_input_schema",
    "build_output_schema",
    "schema",
    "set_type_hints",
    "ValidationError",
    "Discard",
    "ValidatorResult",
    "validate",
    "validator"
]

import apischema.std_types  # Populate custom handlers for stdlib
from .alias import alias
from .conversion import (converter, field_input_converter,
                         field_output_converter, inout_model, input_converter,
                         output_converter)
from .data import from_data, from_stringified, to_data
from .fields import (fields_set, get_fields, mark_set_fields, unmark_set_fields,
                     with_fields_set)
from .ignore import Ignored
from .json_schema import build_input_schema, build_output_schema
from .json_schema.types import JSONSchema
from .properties import properties
from .schema import schema
from .typing import set_type_hints
from .validation.errors import ValidationError
from .validation.validator import Discard, ValidatorResult, validate, validator

del apischema



from dataclasses import (Field, MISSING, _FIELDS,  # type: ignore
                         _FIELD_CLASSVAR, fields, is_dataclass)
from functools import wraps
from inspect import signature
from typing import (AbstractSet, Any, Iterable, Type,
                    TypeVar, cast, overload)

from apischema.utils import PREFIX


class NoDefault(Exception):
    pass


def has_default(field: Field) -> bool:
    return (field.default_factory is not MISSING  # type: ignore
            or field.default is not MISSING)


def get_default(field: Field) -> Any:
    if field.default_factory is not MISSING:  # type: ignore
        return field.default_factory()  # type: ignore
    if field.default is not MISSING:
        return field.default
    raise NoDefault()


FIELDS_SET_ATTR = f"{PREFIX}fields_set"

Cls = TypeVar("Cls", bound=Type)


@overload
def with_fields_set(cls: Cls) -> Cls:
    ...


@overload
def with_fields_set(init: bool) -> Cls:
    ...


def with_fields_set(arg=None, init=True):
    if arg is None:
        return lambda cls: with_fields_set(cls, init)
    cls = arg
    old_init = cls.__init__
    old_setattr = cls.__setattr__
    params = list(signature(old_init).parameters)

    if init:
        @wraps(old_init)
        def new_init(self, *args, **kwargs):
            self.__dict__[FIELDS_SET_ATTR] = set()
            old_init(self, *args, **kwargs)
            self.__dict__[FIELDS_SET_ATTR] = {*params[1:len(args) + 1],
                                              *kwargs}
    else:
        @wraps(old_init)
        def new_init(self, *args, **kwargs):
            self.__dict__[FIELDS_SET_ATTR] = set()
            old_init(self, *args, **kwargs)

    @wraps(old_setattr)
    def new_setattr(self, attr, value):
        self.__dict__[FIELDS_SET_ATTR].add(attr)
        old_setattr(self, attr, value)

    cls.__init__ = new_init
    cls.__setattr__ = new_setattr
    return cls


T = TypeVar("T")


def mark_set_fields(obj: T, *fields: str, overwrite=False) -> T:
    if overwrite:
        obj.__dict__[FIELDS_SET_ATTR] = set(fields)
    else:
        try:
            obj.__dict__[FIELDS_SET_ATTR].update(fields)
        except KeyError:
            fs = fields_set(obj)
            if any(f not in fs for f in fields):
                obj.__dict__[FIELDS_SET_ATTR] = {*fs, *fields}
    return obj


def unmark_set_fields(obj: T, *fields: str) -> T:
    if FIELDS_SET_ATTR in obj.__dict__:
        obj.__dict__[FIELDS_SET_ATTR].difference_update(fields)
    else:
        obj.__dict__[FIELDS_SET_ATTR] = set(fields_set(obj)).difference(fields)
    return obj


def fields_set(obj: Any) -> AbstractSet[str]:
    try:
        return getattr(obj, FIELDS_SET_ATTR)
    except AttributeError:
        if not is_dataclass(obj):
            raise TypeError("`fields_set` can only be called on dataclasses"
                            " and classes decorated with `with_fields_set`")
        return {f.name for f in fields(obj)}


def init_fields(cls: Type) -> Iterable[Field]:
    assert is_dataclass(cls)
    for field in getattr(cls, _FIELDS).values():
        if field.init and field._field_type != _FIELD_CLASSVAR:
            yield field


class FieldGetter:
    def __init__(self, obj):
        assert is_dataclass(obj)
        self.obj = obj
        self.fields = {f.name: f for f in fields(obj)}

    def __getattribute__(self, attr) -> Field:
        try:
            return super().__getattribute__("fields")[attr]
        except KeyError:
            cls = type(super().__getattribute__("obj")).__name__
            raise AttributeError(f"Class {cls} has no field 'attr'")


def get_fields(obj: T) -> T:
    return cast(T, FieldGetter(obj))



from dataclasses import fields
from inspect import isfunction
from typing import Callable, TypeVar, overload

from apischema.types import DictWithUnion, Metadata
from apischema.utils import PREFIX

ALIAS_METADATA = f"{PREFIX}alias"
ALIAS_NO_OVERRIDE_METADATA = f"{PREFIX}alias_no_override"

Cls = TypeVar("Cls")


class ClassAliaser:
    def __init__(self, aliaser: Callable[[str], str]):
        self.aliaser = aliaser

    def __call__(self, cls: Cls) -> Cls:
        for field in fields(cls):
            if field.metadata.get(ALIAS_NO_OVERRIDE_METADATA):
                continue
            alias = self.aliaser(field.metadata.get(ALIAS_METADATA, field.name))
            field.metadata = {**field.metadata, ALIAS_METADATA: alias}
        return cls


@overload
def alias(name: str = None, *, override: bool = True) -> Metadata:
    ...


@overload
def alias(aliaser: Callable[[str], str]) -> ClassAliaser:
    ...


def alias(arg=None, *, override: bool = True):  # type: ignore
    if isfunction(arg):
        return ClassAliaser(arg)
    metadata = {}
    if arg is not None:
        metadata[ALIAS_METADATA] = arg
    if not override:
        metadata[ALIAS_NO_OVERRIDE_METADATA] = True
    if not metadata:
        raise ValueError("Alias must be called with arguments")
    return DictWithUnion(metadata)



__all__ = ["from_data", "from_stringified", "to_data", "build_data"]

from .from_data import from_data
from .from_stringified import from_stringified
from .to_data import to_data
from .utils import build_data



from typing import Any, Dict, Iterable, List, Optional, Tuple, Union


def override_data(obj: Optional[Union[Dict, List]],
                  key: str, value: Any, separator: str) -> Union[Dict, List]:
    assert key
    # Use Any to prevent mypy to complain
    index: Any
    res: Any
    index, remains = key, None
    if separator in key:
        index, remains = key.split(separator, 1)
    if index.isdigit():
        if obj is not None and not isinstance(obj, list):
            raise ValueError(f"expected list at {key}")
        res = obj or []
        index = int(index)
        if index >= len(res):
            res.extend([None] * (index - len(res) + 1))
    else:
        if obj is not None and not isinstance(obj, dict):
            raise ValueError(f"expected dict at {key}")
        res = obj or {}
        res.setdefault(index, None)
    if remains:
        res[index] = override_data(res[index], remains, value, separator)
    else:
        res[index] = value
    return res


def build_data(items: Iterable[Tuple[str, Any]], separator: str = ".") -> Any:
    data = None
    for key, value in items:
        if not key:
            raise ValueError("empty keys are not handled")
        try:
            data = override_data(data, key, value, separator)
        except ValueError as err:
            raise ValueError(f"invalid key '{key}' ({err})")
    return data



from dataclasses import dataclass, field
from enum import Enum
from typing import List, NewType
from uuid import UUID, uuid4

from pytest import raises

from apischema import (ValidationError, build_input_schema, build_output_schema,
                       from_data, schema, to_data)

Tag = NewType("Tag", str)
schema(title="resource tag", max_len=64)(Tag)


class ResourceType(Enum):
    RESOURCE_A = "A"
    RESOURCE_B = "B"


@dataclass
class Resource:
    id: UUID
    type: ResourceType
    tags: List[Tag] = field(default_factory=list,
                            metadata=schema(max_items=5, unique=True))


def test_resource():
    uuid = uuid4()
    data = {
        "id":   str(uuid),
        "type": "A",
        "tags": ["tag1"]
    }
    resource = from_data(data, Resource)
    assert resource == Resource(uuid, ResourceType.RESOURCE_A, [Tag("tag1")])
    assert to_data(resource) == data
    json_schema = build_input_schema(Resource)
    assert json_schema == build_output_schema(Resource)
    assert to_data(json_schema) == {
        "type":                 "object",
        "required":             ["id", "type"],
        "additionalProperties": False,
        "properties":           {
            "id":   {
                "type":   "string",
                "format": "uuid",
            },
            "type": {
                "type": "string",
                "enum": ["A", "B"],
            },
            "tags": {
                "type":        "array",
                "maxItems":    5,
                "uniqueItems": True,
                "items":       {
                    "type":      "string",
                    "title":     "resource tag",
                    "maxLength": 64,
                },
            },
        }
    }


def test_resource_error():
    with raises(ValidationError) as err:
        from_data({"id": "uuid", "type": None, "tags": ["a", "a"]}, Resource)
    assert err.value == ValidationError(children={
        "id":   ValidationError([
            "[ValueError]badly formed hexadecimal UUID string"
        ]),
        "type": ValidationError([
            "None is not a valid ResourceType"
        ]),
        "tags": ValidationError([
            "duplicates items in ['a', 'a'] (uniqueItems)"
        ])
    })



from dataclasses import Field, dataclass, field
from typing import (Any, Dict, Iterable, List, Mapping, Optional, Sequence,
                    Tuple, Union, overload)

from apischema.alias import ALIAS_METADATA

ErrorMsg = str
Error = Union[ErrorMsg, Tuple[Any, ErrorMsg]]

ROOT_KEY = "."


@dataclass
class ValidationError(Exception):
    messages: Sequence[ErrorMsg] = field(default_factory=list)
    children: Mapping[str, "ValidationError"] = field(default_factory=dict)

    def flat(self) -> Iterable[Tuple[Tuple[str, ...], Sequence[ErrorMsg]]]:
        if self.messages:
            yield (), self.messages
        for child_path, child in self.children.items():
            for path, errors in child.flat():
                yield (child_path, *path), errors

    @property
    def flatten(self) -> Mapping[Tuple[str, ...], Sequence[ErrorMsg]]:
        return dict(self.flat())

    def __str__(self):
        return repr(self)


@overload
def merge(err1: Optional[ValidationError], err2: ValidationError
          ) -> ValidationError:
    ...


@overload
def merge(err1: ValidationError, err2: Optional[ValidationError]
          ) -> ValidationError:
    ...


@overload
def merge(err1: Optional[ValidationError], err2: Optional[ValidationError]
          ) -> Optional[ValidationError]:
    ...


def merge(err1: Optional[ValidationError], err2: Optional[ValidationError]):
    if err1 is None:
        return err2
    if err2 is None:
        return err1
    return ValidationError([*err1.messages, *err2.messages], {
        key: merge(err1.children.get(key), err2.children.get(key))  # type: ignore
        for key in err1.children.keys() | err2.children.keys()
    })


def exception(err: Exception) -> str:
    return f"[{type(err).__name__}]{err}"


def _rec_build_error(path: Sequence[str], msg: ErrorMsg) -> ValidationError:
    if not path:
        return ValidationError([msg])
    else:
        return ValidationError(
            children={path[0]: _rec_build_error(path[1:], msg)})


def _check_error_path(path: Any) -> Sequence[str]:
    if isinstance(path, (Field, int, str)):
        path = [path]
    else:
        path = list(path)
    for i, elt in enumerate(path):
        if isinstance(elt, Field):
            path[i] = elt.metadata.get(ALIAS_METADATA, elt.name)
        elif isinstance(elt, int):
            path[i] = str(elt)
        if not isinstance(path[i], str):
            raise TypeError(f"Bad error path, expected Field, int or str,"
                            f" found {type(i)}")
    return path


def build_from_errors(errors: Iterable[Error]) -> ValidationError:
    messages: List[ErrorMsg] = []
    children: Dict[str, ValidationError] = {}
    for error in errors:
        if isinstance(error, ErrorMsg):
            messages.append(error)
            continue
        path, msg = error
        path = _check_error_path(path)
        if not path:
            messages.append(msg)
        else:
            children[path[0]] = merge(children.get(path[0]),
                                      _rec_build_error(path[1:], msg))
    return ValidationError(messages, children)



from pytest import raises

from apischema.utils import distinct, to_generator


def test_make_gen():
    def func():
        return 42

    with raises(StopIteration) as err:
        next(iter(to_generator(func)()))
    assert err.value.value == 42


def test_distinct():
    assert list(distinct([4, 0, 1, 0, 4])) == [4, 0, 1]



from dataclasses import dataclass
from typing import NewType

from apischema import (from_data, input_converter, output_converter, schema,
                       to_data)

HexaRGB = NewType("HexaRGB", str)
schema(pattern="^#[0-9a-fA-F]{6}$")(HexaRGB)


@dataclass
class RGB:
    red: int
    green: int
    blue: int

    @output_converter
    def hexa(self) -> HexaRGB:
        return HexaRGB(f"#{self.red:02x}{self.green:02x}{self.blue:02x}")


@input_converter
def rgb_from_str(hexa: HexaRGB) -> RGB:
    return RGB(int(hexa[1:3], 16), int(hexa[3:5], 16), int(hexa[5:7], 16))


def test_rgb():
    assert from_data("#FFFFFF", RGB) == RGB(255, 255, 255)
    assert to_data(RGB(0, 0, 0)) == "#000000"

# Standard library types like UUID or datetime are handled this way
# see std_types.py



from dataclasses import (Field, InitVar, _FIELD_INITVAR, fields,  # type: ignore
                         is_dataclass, replace)
from enum import Enum
from itertools import chain
from typing import (Any, Callable, Dict, Iterable, List, Mapping, Optional, Sequence,
                    Set, Tuple, Type, Union)

from apischema.alias import ALIAS_METADATA
from apischema.conversion import (Converter, INPUT_METADATA, InputVisitorMixin,
                                  OUTPUT_METADATA, OutputVisitorMixin)
from apischema.data import to_data
from apischema.fields import (NoDefault, fields_set, get_default, has_default,
                              init_fields,
                              mark_set_fields)
from apischema.ignore import Ignored
from apischema.json_schema.types import JSONSchema, JSONType
from apischema.properties import PROPERTIES_METADATA
from apischema.schema import Constraint, Schema
from apischema.schema.annotations import ANNOTATIONS_METADATA, get_annotations
from apischema.schema.constraints import (ArrayConstraint, CONSTRAINT_METADATA,
                                          NumberConstraint, ObjectConstraint,
                                          StringConstraint, get_constraint)
from apischema.typing import get_type_hints
from apischema.utils import NO_DEFAULT, as_dict
from apischema.visitor import Visitor

constraint_by_type = {
    int:   NumberConstraint,
    float: NumberConstraint,
    str:   StringConstraint,
}


class ReadWriteOnly(Constraint):
    pass


def check_constraint(schema: Schema, expected: Type[Constraint]):
    if schema.constraint is not None:
        if not isinstance(schema.constraint, expected):
            if not isinstance(schema.constraint, ReadWriteOnly):
                raise TypeError(f"Bad constraint: expected {expected.__name__}"
                                f" found {type(schema.constraint).__name__}")


def _remove_none(obj: Optional[Any]) -> Iterable[Tuple[str, Any]]:
    if obj is None:
        return iter(())
    return ((k, v) for k, v in as_dict(obj).items()
            if v is not None and v is not NO_DEFAULT
            and k not in ("items", "additional_properties"))


def _to_dict(schema: Schema) -> Mapping[str, Any]:
    return dict(chain(_remove_none(schema.annotations),
                      _remove_none(schema.constraint)))


def _merge_constraints(c1: Optional[Constraint],
                       c2: Optional[Constraint]) -> Optional[Constraint]:
    if c1 is None:
        return c2
    if c2 is None:
        return c1
    if isinstance(c1, ReadWriteOnly):
        return type(c2)(**{**as_dict(c1), **as_dict(c2)})
    elif Constraint != type(c1) != type(c2) != Constraint:
        raise TypeError(f"Incompatibles constraints {type(c1).__name__}"
                        f" and {type(c2).__name__}")
    return c1


def _override(schema: Schema, cls: Type) -> Schema:
    return Schema(schema.annotations or get_annotations(cls),
                  _merge_constraints(schema.constraint, get_constraint(cls)))


def _field_schema(field: Field) -> Schema:
    return Schema(field.metadata.get(ANNOTATIONS_METADATA),
                  field.metadata.get(CONSTRAINT_METADATA))


class SchemaBuilder(Visitor[Schema, JSONSchema]):
    def __init__(self, ref_factory: Optional[Callable[[Type], str]]):
        super().__init__()
        self.ref_factory = ref_factory or self._missing_ref_factory
        self.schemas: Set[Type] = set()

    @staticmethod
    def _missing_ref_factory(cls: Type) -> str:
        raise TypeError("reference factory is needed to handle"
                        " recursive types")

    def primitive(self, cls: Type, schema: Schema) -> JSONSchema:
        if cls in constraint_by_type:
            check_constraint(schema, constraint_by_type[cls])
        return JSONSchema(type=JSONType.from_type(cls), **_to_dict(schema))

    def union(self, alternatives: Sequence[Type],
              schema: Schema) -> JSONSchema:
        any_of = []
        for cls in alternatives:
            try:
                any_of.append(self.visit(cls, schema))
            except Ignored:
                pass
        if len(any_of) == 1:
            return any_of[0]
        else:
            return JSONSchema(any_of=any_of)

    def iterable(self, cls: Type[Iterable], value_type: Type,
                 schema: Schema) -> JSONSchema:
        check_constraint(schema, ArrayConstraint)
        if isinstance(schema.items_, Sequence):
            raise TypeError("Iterable items schema cannot be a list of schema")
        return JSONSchema(type=JSONType.ARRAY,
                          items=self.visit(value_type, schema.items_),
                          **_to_dict(schema))

    def mapping(self, cls: Type[Mapping], key_type: Type, value_type: Type,
                schema: Schema) -> JSONSchema:
        check_constraint(schema, ObjectConstraint)
        props = self.visit(value_type, schema.additional_properties)
        return JSONSchema(type=JSONType.OBJECT, additional_properties=props,
                          **_to_dict(schema))

    def typed_dict(self, cls: Type, keys: Mapping[str, Type], total: bool,
                   schema: Schema) -> JSONSchema:
        if cls in self.schemas:
            return JSONSchema(ref=self.ref_factory(cls))
        check_constraint(schema, ObjectConstraint)
        if schema.additional_properties:
            raise TypeError("additional properties are not handled"
                            " for TypedDict")
        types = get_type_hints(cls)
        sorted_keys = sorted(keys)
        return JSONSchema(type=JSONType.OBJECT,
                          properties={key: self.visit(types[key], Schema())
                                      for key in sorted_keys},
                          required=sorted_keys if total else [],
                          additional_properties=True, **_to_dict(schema))

    def tuple(self, types: Sequence[Type], schema: Schema) -> JSONSchema:
        check_constraint(schema, ArrayConstraint)
        if schema.constraint is not None:
            assert isinstance(schema.constraint, ArrayConstraint)
            if (schema.constraint.max_items is not None
                    or schema.constraint.min_items is not None):
                raise TypeError("Tuple cannot have min_items/max_items"
                                " constraint")
        if isinstance(schema.items_, Sequence):
            if len(schema.items_) != len(types):
                raise TypeError("Tuple items schema list must have the same"
                                " length than tuple")
        items = []
        for i, cls in enumerate(types):
            if isinstance(schema.items_, Sequence):
                elt_schema = schema.items_[i]
            else:
                elt_schema = schema.items_
            items.append(self.visit(cls, elt_schema))
        return JSONSchema(type=JSONType.ARRAY, items=items,
                          min_items=len(types), max_items=len(types),
                          **_to_dict(schema))

    def literal(self, values: Sequence[Any], schema: Schema) -> JSONSchema:
        if not values:
            raise TypeError("empty Literal")
        if schema.constraint is not None:
            raise TypeError("Literal cannot have constraint")
        types = sorted(set(
            JSONType.from_type(type(v.value if isinstance(v, Enum) else v))
            for v in values
        ))
        type_ = types[0] if len(types) == 1 else types
        if len(values) == 1:
            return JSONSchema(type=type_, const=values[0],  # type: ignore
                              **_to_dict(schema))
        else:
            return JSONSchema(type=type_, enum=values,  # type: ignore
                              **_to_dict(schema))

    def _dataclass_fields(self, cls: Type) -> Iterable[Field]:
        raise NotImplementedError()

    def _set_field_default(self, field: Field, schema: Schema):
        if schema.annotations is not None and schema.annotations.default is ...:
            try:
                default = to_data(get_default(field))
            except NoDefault:
                raise TypeError("Invalid ... as default without field default")
            schema.annotations = replace(schema.annotations, default=default)

    def _field_visit(self, field: Field, types: Mapping[str, Type]
                     ) -> JSONSchema:
        raise NotImplementedError()

    def dataclass(self, cls: Type, schema: Schema) -> JSONSchema:
        assert is_dataclass(cls)
        if cls in self.schemas:
            return JSONSchema(ref=self.ref_factory(cls))
        self.schemas.add(cls)
        check_constraint(schema, ObjectConstraint)
        schema_ = _override(schema, cls)
        properties = {}
        additional_field: Optional[Field] = None
        pattern_fields: List[Field] = []
        types = get_type_hints(cls)
        required: List[str] = []
        for field in self._dataclass_fields(cls):
            if PROPERTIES_METADATA in field.metadata:
                if field.metadata[PROPERTIES_METADATA] is None:
                    if additional_field is not None:
                        raise TypeError("Multiple properties without pattern")
                    additional_field = field
                else:
                    pattern_fields.append(field)
                continue
            alias = field.metadata.get(ALIAS_METADATA, field.name)
            properties[alias] = self._field_visit(field, types)
            if not has_default(field):
                required.append(alias)
        pattern_properties: Dict[str, Any] = {
            field.metadata[PROPERTIES_METADATA].pattern: self._field_visit(field, types)
            for field in pattern_fields
        }
        additional_properties: Optional[Union[bool, "JSONSchema"]] = False
        if additional_field:
            additional_properties = self._field_visit(additional_field, types)
        others: Dict[str, Any] = {}
        if required:
            others["required"] = required
        if pattern_properties:
            others["pattern_properties"] = pattern_properties
        return JSONSchema(type=JSONType.OBJECT,  # type: ignore
                          properties=properties,
                          additional_properties=additional_properties,
                          **others, **_to_dict(schema_))

    def enum(self, cls: Type[Enum], schema: Schema) -> JSONSchema:
        if schema.constraint is not None:
            raise TypeError("enum cannot have constraint")
        if len(cls) == 0:
            raise TypeError("Empty enum")
        return self.literal(list(cls), _override(schema, cls))

    def new_type(self, cls: Type, super_type: Type, schema: Schema
                 ) -> JSONSchema:
        return self.visit(super_type, _override(schema, cls))

    def any(self, schema: Schema) -> JSONSchema:
        if schema.constraint is not None:
            raise TypeError("Any type cannot have constraint")
        return JSONSchema()

    def annotated(self, cls: Type, annotations: Sequence[Any],
                  schema: Schema) -> JSONSchema:
        schema_ = schema
        if Ignored in annotations:
            raise Ignored
        for annotation in annotations:
            if isinstance(annotation, Schema):
                schema_ = Schema(
                    schema.annotations or annotation.annotations,
                    _merge_constraints(schema.constraint,
                                       annotation.constraint)
                )
                break
        return self.visit(cls, schema_)


class InputSchemaBuilder(InputVisitorMixin[Schema, JSONSchema], SchemaBuilder):
    def __init__(self, ref_factory: Optional[Callable[[Type], str]],
                 additional_properties: bool):
        SchemaBuilder.__init__(self, ref_factory)
        self.additional_properties = additional_properties

    def _custom(self, cls: Type, custom: Dict[Type, Converter],
                schema: Schema) -> JSONSchema:
        return self.union(list(custom), _override(schema, cls))

    def _dataclass_fields(self, cls: Type) -> Iterable[Field]:
        return init_fields(cls)

    def _field_visit(self, field: Field, types: Mapping[str, Type]
                     ) -> JSONSchema:
        if INPUT_METADATA in field.metadata:
            cls, _ = field.metadata[INPUT_METADATA]
        else:
            cls = types[field.name]
        if isinstance(cls, InitVar):
            cls = cls.type  # type: ignore
        schema = _field_schema(field)
        self._set_field_default(field, schema)
        if field._field_type == _FIELD_INITVAR:  # type: ignore
            if schema.constraint is None:
                schema.constraint = ReadWriteOnly(write_only=True)
            else:
                if schema.constraint.read_only:
                    raise TypeError("InitVar cannot be read-only")
                schema.constraint = replace(schema.constraint, write_only=True)
        return self.visit(cls, schema)

    def dataclass(self, cls: Type, schema: Schema) -> JSONSchema:
        json_schema = super().dataclass(cls, schema)
        if json_schema.additional_properties is False:
            # TODO document that `dataclasses.replace` don't keep fields set
            return mark_set_fields(replace(
                json_schema, additional_properties=self.additional_properties
            ), *fields_set(json_schema), overwrite=True)
        return json_schema


def build_input_schema(cls: Type, *, ref_factory: Callable[[Type], str] = None,
                       additional_properties: bool = False,
                       schema: Schema = None) -> JSONSchema:
    builder = InputSchemaBuilder(ref_factory, additional_properties)
    return builder.visit(cls, schema or Schema())


class OutputSchemaBuilder(OutputVisitorMixin[Schema, JSONSchema],
                          SchemaBuilder):
    def __init__(self, conversions: Mapping[Type, Type],
                 ref_factory: Optional[Callable[[Type], str]]):
        SchemaBuilder.__init__(self, ref_factory)
        OutputVisitorMixin.__init__(self, conversions)

    def _custom(self, cls: Type, custom: Tuple[Type, Converter],
                schema: Schema) -> JSONSchema:
        return self.visit(custom[0], _override(schema, cls))

    def _dataclass_fields(self, cls: Type) -> Iterable[Field]:
        assert is_dataclass(cls)
        return fields(cls)

    def _field_visit(self, field: Field, types: Mapping[str, Type]
                     ) -> JSONSchema:
        if OUTPUT_METADATA in field.metadata:
            cls, _ = field.metadata[OUTPUT_METADATA]
        else:
            cls = types[field.name]
        schema = _field_schema(field)
        self._set_field_default(field, schema)
        if not field.init:
            if schema.constraint is None:
                schema.constraint = ReadWriteOnly(read_only=True)
            else:
                if schema.constraint.write_only:
                    raise TypeError("not init field cannot be write-only")
                schema.constraint = replace(schema.constraint, read_only=True)
        return self.visit(cls, schema)


def build_output_schema(cls: Type, conversions: Mapping[Type, Type] = None, *,
                        ref_factory: Callable[[Type], str] = None,
                        schema: Schema = None) -> JSONSchema:
    builder = OutputSchemaBuilder(conversions or {}, ref_factory)
    return builder.visit(cls, schema or Schema())



from dataclasses import fields, is_dataclass
from functools import wraps
from typing import (Any, Callable, Dict, Generator, Generic, Hashable, Iterable,
                    TypeVar,
                    Union)

PREFIX = "_apischema_"
NO_DEFAULT = object()

Func = TypeVar("Func", bound=Callable)


def to_generator(func: Callable):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
        yield  # noqa

    return wrapper


Hashable_ = TypeVar("Hashable_", bound=Hashable)


def distinct(values: Iterable[Hashable_]) -> Iterable[Hashable_]:
    unique = set()
    for value in values:
        if value not in unique:
            unique.add(value)
            yield value


def to_hashable(data: Union[None, int, float, str, bool, list, dict]
                ) -> Hashable:
    if isinstance(data, list):
        return tuple(map(to_hashable, data))
    if isinstance(data, dict):
        return tuple(sorted((to_hashable(k), to_hashable(v))
                            for k, v in data.items()))
    return data  # type: ignore


T = TypeVar("T")
U = TypeVar("U")


class GeneratorValue(Iterable[U], Generic[T, U]):
    value: T

    def __init__(self, generator: Generator[U, None, T]):
        self.generator = generator

    def __iter__(self):
        self.value = yield from self.generator


def to_camel_case(s: str):
    pascal_case = "".join(map(str.capitalize, s.split("_")))
    return pascal_case[0].lower() + pascal_case[1:]


def as_dict(obj) -> Dict[str, Any]:
    assert is_dataclass(obj)
    return {f.name: getattr(obj, f.name) for f in fields(obj)}



from dataclasses import is_dataclass
from enum import Enum, EnumMeta
from typing import (Any, Dict, Generic, Iterable, Mapping, Sequence, Type,
                    TypeVar, Union)

from apischema.types import ITERABLE_TYPES, MAPPING_TYPES, PRIMITIVE_TYPE
from apischema.typing import (Literal, NamedTupleMeta, _AnnotatedAlias,
                              _GenericAlias, _TypedDictMeta, _type_repr,
                              get_type_hints)

PRIMITIVE_TYPE_IDS = set(map(id, PRIMITIVE_TYPE))


class Unsupported(TypeError):
    def __init__(self, cls: Type):
        self.cls = cls

    def __repr__(self):
        return f"unsupported '{_type_repr(self.cls)}' type"


class NotCustom:
    pass


NOT_CUSTOM = NotCustom()

Arg = TypeVar("Arg", contravariant=True)
Return = TypeVar("Return", covariant=True)


class Visitor(Generic[Arg, Return]):
    def __init__(self):
        self._generics: Dict[TypeVar, Type] = {}

    def primitive(self, cls: Type, arg: Arg) -> Return:
        raise NotImplementedError()

    def union(self, alternatives: Sequence[Type], arg: Arg) -> Return:
        raise NotImplementedError()

    def iterable(self, cls: Type[Iterable], value_type: Type,
                 arg: Arg) -> Return:
        raise NotImplementedError()

    def mapping(self, cls: Type[Mapping], key_type: Type, value_type: Type,
                arg: Arg) -> Return:
        raise NotImplementedError()

    def typed_dict(self, cls: Type, keys: Mapping[str, Type], total: bool,
                   arg: Arg) -> Return:
        raise NotImplementedError()

    def tuple(self, types: Sequence[Type], arg: Arg) -> Return:
        raise NotImplementedError()

    def literal(self, values: Sequence[Any], arg: Arg) -> Return:
        raise NotImplementedError()

    def custom(self, cls: Type, arg: Arg) -> Union[Return, NotCustom]:
        return NOT_CUSTOM

    def dataclass(self, cls: Type, arg: Arg) -> Return:
        raise NotImplementedError()

    def enum(self, cls: Type[Enum], arg: Arg) -> Return:
        raise NotImplementedError()

    def new_type(self, cls: Type, super_type: Type, arg: Arg) -> Return:
        return self.visit(super_type, arg)

    def any(self, arg: Arg) -> Return:
        raise NotImplementedError()

    def annotated(self, cls: Type, annotations: Sequence[Any],
                  arg: Arg) -> Return:
        return self.visit(cls, arg)

    def named_tuple(self, cls: Type, types: Mapping[str, Type],
                    defaults: Mapping[str, Any], arg: Arg) -> Return:
        raise TypeError("NamedTuple is not handled")

    def visit(self, cls: Type, arg: Arg) -> Return:
        # Use `id` to avoid useless costly generic types hashing
        if id(cls) in PRIMITIVE_TYPE_IDS:
            return self.primitive(cls, arg)
        if isinstance(cls, _GenericAlias):
            origin = cls.__origin__
            if isinstance(cls, _AnnotatedAlias):
                return self.annotated(origin, cls.__metadata__, arg)
            if origin is Union:
                return self.union(cls.__args__, arg)
            if origin is tuple:
                if len(cls.__args__) < 2 or cls.__args__[1] is not ...:
                    return self.tuple(cls.__args__, arg)
            if origin in ITERABLE_TYPES:
                return self.iterable(origin, cls.__args__[0], arg)
            if origin in MAPPING_TYPES:
                return self.mapping(origin, cls.__args__[0], cls.__args__[1],
                                    arg)
            if origin is Literal:
                return self.literal(cls.__args__, arg)
            custom = self.custom(cls, arg)
            if custom is not NOT_CUSTOM:
                assert not isinstance(custom, NotCustom)
                return custom
            # TypeVar handling
            generics_save = self._generics.copy()
            for tv, value in zip(origin.__parameters__, cls.__args__):
                # Kind of ugly side effect, but for simplicity
                self._generics[tv] = generics_save.get(value, value)
            res = self.visit(origin, arg)
            self._generics = generics_save
            return res
        # customs are handled before other classes
        custom = self.custom(cls, arg)
        if custom is not NOT_CUSTOM:
            assert not isinstance(custom, NotCustom)
            return custom
        if is_dataclass(cls):
            return self.dataclass(cls, arg)
        if isinstance(cls, _TypedDictMeta):
            return self.typed_dict(cls, get_type_hints(cls),
                                   cls.__total__, arg)  # type: ignore
        if isinstance(cls, TypeVar):  # type: ignore
            try:
                cls_ = self._generics[cls]
            except KeyError:
                if cls.__constraints__:
                    return self.visit(Union[cls.__constraints__], arg)
                return self.visit(Any, arg)  # type: ignore
            else:
                return self.visit(cls_, arg)
        if isinstance(cls, EnumMeta):
            return self.enum(cls, arg)
        if hasattr(cls, "__supertype__"):
            return self.new_type(cls, cls.__supertype__, arg)
        if cls is Any:
            return self.any(arg)
        if isinstance(cls, NamedTupleMeta):
            if hasattr(cls, "__annotations__"):
                types = get_type_hints(cls)
            else:
                types = cls._field_types  # type: ignore
            return self.named_tuple(cls, types, cls._field_defaults, arg)
        if hasattr(cls, "__parameters__"):
            params = tuple(self._generics.get(p, Any)
                           for p in getattr(cls, "__parameters__"))
            raise Unsupported(cls[params])
        raise Unsupported(cls)



from dataclasses import Field, dataclass
from functools import partial
from types import FunctionType, MethodType
from typing import Any, Callable, ClassVar, Mapping, Optional, Type

from apischema.fields import FIELDS_SET_ATTR, get_default
from apischema.typing import get_type_hints

MOCK_FIELDS_FIELD = "__mock_fields__"
MOCK_CLS_FIELD = "__mock_cls__"


class NonTrivialDependency(Exception):
    def __init__(self, attr: str):
        self.attr = attr
        self.validator: Optional[Callable] = None


@dataclass
class ValidatorMock:
    def __init__(self, cls: Type, fields: Mapping[str, Any],
                 defaults: Mapping[str, Field]):
        self.cls = cls
        self.fields = fields
        self.defaults = defaults
        setattr(self, FIELDS_SET_ATTR, set(fields))

    def __getattribute__(self, attr: str) -> Any:
        if attr in super().__getattribute__("fields"):
            return super().__getattribute__("fields")[attr]
        if attr in super().__getattribute__("defaults"):
            return get_default(super().__getattribute__("defaults")[attr])
        if attr == "__class__":
            return super().__getattribute__("cls")
        if attr == "__dict__":
            return {**super().__getattribute__("fields"),
                    **{name: get_default(field)
                       for name, field
                       in super().__getattribute__("defaults").items()},
                    FIELDS_SET_ATTR: set(super().__getattribute__("fields"))}
        if hasattr(super().__getattribute__("cls"), attr):
            member = getattr(super().__getattribute__("cls"), attr)
            # for classmethod (staticmethod are not handled)
            if isinstance(member, MethodType):
                return member
            if isinstance(member, FunctionType):
                return partial(member, self)
            if isinstance(member, property):
                return member.fget(self)  # type: ignore
            types = get_type_hints(super().__getattribute__("cls"))
            if getattr(types.get(attr), "__origin__",
                       types.get(attr)) is ClassVar:
                return member
        raise NonTrivialDependency(attr)



from enum import Enum
from itertools import chain
from typing import Dict, Iterable, Tuple, Type, TypeVar

from apischema.data.from_data import DataWithConstraint, FromData
from apischema.data.utils import override_data

_bool_pairs = (
    ("0", "1"),
    ("f", "t"),
    ("n", "y"),
    ("no", "yes"),
    ("false", "true"),
    ("off", "on"),
    ("ko", "ok"),
)
STR_TO_BOOL: Dict[str, bool] = {}
for false, true in _bool_pairs:
    for s, value in ((false, False), (true, True)):
        STR_TO_BOOL[s] = value
        STR_TO_BOOL[s.capitalize()] = value
        STR_TO_BOOL[s.upper()] = value

STR_NONE_VALUES = set(chain.from_iterable(
    (s, s.capitalize(), s.upper()) for s in ("", "null", "none", "nil")
))


class FromStringified(FromData):
    def primitive(self, cls, data2: DataWithConstraint):
        data, constraint = data2
        if not isinstance(data, str):
            raise ValueError("stringified values must be string")
        if cls is type(None) and data in STR_NONE_VALUES:  # type: ignore # noqa
            return None
        if cls is bool and data in STR_TO_BOOL:
            return STR_TO_BOOL[data]
        try:
            return cls(data)
        except ValueError:
            return super().primitive(cls, data2)

    def enum(self, cls: Type[Enum], data2: DataWithConstraint):
        data, constraint = data2
        data = self.primitive(type(data), (data, None))
        return super().enum(cls, (data, constraint))


T = TypeVar("T")


def from_stringified(items: Iterable[Tuple[str, str]], cls: Type[T], *,
                     additional_properties: bool = False,
                     separator: str = ".") -> T:
    data = None
    for key, value in items:
        if not key:
            raise ValueError("empty keys are not handled")
        try:
            data = override_data(data, key, value, separator)
        except ValueError:
            raise ValueError(f"invalid key '{key}'")
    visitor = FromStringified(additional_properties)
    return visitor.visit(cls, (data, None))



__all__ = ["schema", "Schema", "Annotations", "Constraint"]

from dataclasses import dataclass, fields
from typing import (Any, Dict, Optional, Sequence, TypeVar,
                    Union, overload)

from .annotations import ANNOTATIONS_METADATA, Annotations, _annotations
from .constraints import (ArrayConstraint, CONSTRAINT_METADATA, Constraint,
                          NumberConstraint,
                          ObjectConstraint, StringConstraint, _constraints)
from ..types import MetadataMixin, Number

T = TypeVar("T")


@dataclass
class Schema(MetadataMixin):
    annotations: Optional[Annotations] = None
    constraint: Optional[Constraint] = None

    def __post_init__(self):
        self.metadata: Dict[str, Any] = {}
        if self.annotations is not None:
            self.metadata[ANNOTATIONS_METADATA] = self.annotations
        if self.constraint is not None:
            self.metadata[CONSTRAINT_METADATA] = self.constraint

    def __call__(self, obj: T) -> T:
        if self.annotations is not None:
            _annotations[obj] = self.annotations
        if self.constraint is not None:
            _constraints[obj] = self.constraint
        return obj

    @property
    def items_(self) -> Union["Schema", Sequence["Schema"]]:
        assert (self.constraint is None
                or isinstance(self.constraint, ArrayConstraint))
        annotations, constraint = None, None
        if self.annotations is not None:
            annotations = self.annotations.items
        if self.constraint is not None:
            constraint = self.constraint.items
        if isinstance(annotations, Sequence):
            assert isinstance(constraint, Sequence)
            assert len(annotations) == len(constraint)
            return [Schema(a, c) for a, c in zip(annotations, constraint)]
        else:
            assert annotations is None or isinstance(annotations, Annotations)
            assert constraint is None or isinstance(constraint, Constraint)
            return Schema(annotations, constraint)

    @property
    def additional_properties(self) -> "Schema":
        assert (self.constraint is None
                or isinstance(self.constraint, ObjectConstraint))
        annotations, constraint = None, None
        if self.annotations is not None:
            annotations = self.annotations.additional_properties
        if self.constraint is not None:
            constraint = self.constraint.additional_properties
        return Schema(annotations, constraint)


_constraint_rewrite = {
    "min":        "minimum",
    "max":        "maximum",
    "exc_min":    "exclusive_minimum",
    "exc_max":    "exclusive_maximum",
    "mult_of":    "multiple_of",
    "min_len":    "min_length",
    "max_len":    "max_length",
    "unique":     "unique_items",
    "properties": "additional_properties",
}


@overload
def schema(*, title: Optional[str] = None,
           description: Optional[str] = None,
           examples: Optional[Sequence[Any]] = None,
           read_only: Optional[bool] = None,
           write_only: Optional[bool] = None) -> Schema:
    ...


@overload
def schema(*, title: Optional[str] = None,
           description: Optional[str] = None,
           examples: Optional[Sequence[Any]] = None,
           read_only: Optional[bool] = None,
           write_only: Optional[bool] = None,
           min: Optional[Number] = None,
           max: Optional[Number] = None,
           exc_min: Optional[Number] = None,
           exc_max: Optional[Number] = None,
           mult_of: Optional[Number] = None) -> Schema:
    ...


@overload
def schema(*, title: Optional[str] = None,
           description: Optional[str] = None,
           examples: Optional[Sequence[Any]] = None,
           read_only: Optional[bool] = None,
           write_only: Optional[bool] = None,
           format: Optional[str] = None,
           min_len: Optional[int] = None,
           max_len: Optional[int] = None,
           pattern: Optional[str] = None) -> Schema:
    ...


@overload
def schema(*, title: Optional[str] = None,
           description: Optional[str] = None,
           examples: Optional[Sequence[Any]] = None,
           read_only: Optional[bool] = None,
           write_only: Optional[bool] = None,
           min_items: Optional[int] = None,
           max_items: Optional[int] = None,
           unique: Optional[bool] = None,
           items: Optional[Union[Schema, Sequence[Schema]]] = None) -> Schema:
    ...


@overload
def schema(*, title: Optional[str] = None,
           description: Optional[str] = None,
           examples: Optional[Sequence[Any]] = None,
           read_only: Optional[bool] = None,
           write_only: Optional[bool] = None,
           min_properties: Optional[int] = None,
           max_properties: Optional[int] = None,
           properties: Optional[Schema] = None) -> Schema:
    ...


def schema(**kwargs) -> Schema:
    kwargs_ = {_constraint_rewrite.get(k, k): v
               for k, v in kwargs.items()}
    annotations_fields = {f.name for f in fields(Annotations)}
    annotations_kwargs = {k: v for k, v in kwargs_.items()
                          if k in annotations_fields}
    constraint_kwargs = {k: v for k, v in kwargs_.items()
                         if k not in annotations_kwargs}
    if "additional_properties" in annotations_kwargs:
        schema1: Schema = annotations_kwargs["additional_properties"]
        annotations_kwargs["additional_properties"] = schema1.annotations
        constraint_kwargs["additional_properties"] = schema1.constraint
    if "items" in annotations_kwargs:
        schema2: Union[Schema, Sequence[Schema]] = annotations_kwargs["items"]
        if not isinstance(schema2, Sequence):
            annotations_kwargs["items"] = [s.annotations for s in schema2]
            constraint_kwargs["items"] = [s.constraint for s in schema2]
        else:
            assert isinstance(schema2, Schema)
            annotations_kwargs["items"] = schema2.annotations
            constraint_kwargs["items"] = schema2.constraint

    annotations = None
    if annotations_kwargs:
        annotations = Annotations(**annotations_kwargs)

    constraint = None
    if constraint_kwargs:
        for cls in (NumberConstraint, StringConstraint,
                    ArrayConstraint, ObjectConstraint):
            try:
                constraint = cls(**constraint_kwargs)
                break
            except TypeError:
                pass

    return Schema(annotations, constraint)



import collections.abc
import sys
from typing import (AbstractSet, Any, Collection, Dict, Iterable, List, Mapping,
                    MutableSequence, Sequence, Set, Tuple, Union)

Number = Union[int, float]

PRIMITIVE_TYPE = {str, int, bool, float, type(None)}

# Hack before PEP 585 ...
ITERABLE_TYPES = {
    collections.abc.Iterable:        tuple,
    collections.abc.Collection:      tuple,
    collections.abc.Sequence:        tuple,
    tuple:                           tuple,
    collections.abc.MutableSequence: list,
    list:                            list,
    collections.abc.Set:             frozenset,
    frozenset:                       frozenset,
    collections.abc.MutableSet:      set,
    set:                             set
}

MAPPING_TYPES = {
    collections.abc.Mapping:        dict,
    collections.abc.MutableMapping: dict,
    dict:                           dict
}

UNTYPED_COLLECTIONS = {
    tuple:     Tuple[Any, ...],
    list:      List[Any],
    frozenset: AbstractSet[Any],
    set:       Set[Any],
    dict:      Dict[Any, Any]
}

TYPED_ORIGINS = {
    tuple:                           Tuple,
    list:                            List,
    frozenset:                       AbstractSet,
    set:                             Set,
    dict:                            Dict,
    collections.abc.Iterable:        Iterable,
    collections.abc.Collection:      Collection,
    collections.abc.Sequence:        Sequence,
    collections.abc.MutableSequence: MutableSequence,
    collections.abc.Set:             AbstractSet,
    collections.abc.MutableSet:      Set,
}

if sys.version_info >= (3, 7):
    OrderedDict = dict
else:
    from collections import OrderedDict  # noqa

# Kind of hack to benefit of PEP 584
if sys.version_info >= (3, 9):
    Metadata = Mapping[str, Any]
    DictWithUnion = dict
else:
    class Metadata(Mapping[str, Any]):
        def __or__(self, other: Mapping[str, Any]) -> "Metadata":
            return DictWithUnion(**self, **other)


    class DictWithUnion(Dict[str, Any], Metadata):  # noqa
        pass


class MetadataMixin(Metadata):
    metadata: Mapping[str, Any]

    def __getitem__(self, key):
        return self.metadata[key]

    def __iter__(self):
        return iter(self.metadata)

    def __len__(self):
        return len(self.metadata)



from collections import defaultdict
from dataclasses import (MISSING, dataclass, field, fields, is_dataclass)
from inspect import (Parameter, isclass, isgeneratorfunction,
                     signature)
from logging import getLogger
from types import MappingProxyType, MethodType
from typing import (Any, Callable, Dict, Generator, Generic, List, Mapping,
                    Optional, Tuple, Type, TypeVar, Union, cast,
                    get_type_hints)

from apischema.properties import properties
from apischema.types import (DictWithUnion, Metadata, TYPED_ORIGINS)
from apischema.typing import _GenericAlias
from apischema.utils import GeneratorValue, PREFIX, as_dict, to_camel_case
from apischema.validation.errors import Error
from apischema.visitor import NOT_CUSTOM, NotCustom

logger = getLogger(__name__)

Converter = Callable[[Any], Any]

_input_converters: Dict[Type, Dict[Type, Converter]] = defaultdict(dict)
_output_converters: Dict[Type, Tuple[Type, Converter]] = {}
_converters: Dict[Type, Dict[Type, Converter]] = defaultdict(dict)

Param = TypeVar("Param")
Return = TypeVar("Return")


def check_converter(converter: Converter,
                    param: Optional[Type[Param]],
                    ret: Optional[Type[Return]]
                    ) -> Tuple[Type[Param], Type[Return]]:
    try:
        parameters = iter(signature(converter).parameters.values())
    except ValueError:  # builtin types
        if ret is None and isclass(ret):
            ret = cast(Type[Any], converter)
        if param is None:
            raise TypeError("converter parameter must be typed")
    else:
        try:
            first = next(parameters)
        except StopIteration:
            raise TypeError("converter must have at least one parameter")
        for p in parameters:
            if p.default is Parameter.empty and p.kind not in (
                    Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD):
                raise TypeError("converter must have at most one parameter "
                                "without default")
        types = get_type_hints(converter)
        if param is None:
            try:
                param = types.pop(first.name)
            except KeyError:
                raise TypeError("converter parameter must be typed")
        if ret is None:
            try:
                ret = types.pop("return")
            except KeyError:
                raise TypeError("converter return must be typed")
    return cast(Type[Param], param), cast(Type[Return], ret)


class ConversionError(Exception):
    def __init__(self, errors: List[Error]):
        self.errors = errors


def handle_potential_validation(ret: Type, converter: Converter
                                ) -> Tuple[Type, Converter]:
    if not isgeneratorfunction(converter):
        return ret, converter

    def wrapper(arg):
        generator = GeneratorValue(converter(arg))
        errors = list(generator)
        if errors:
            raise ConversionError(errors)
        return generator.value

    if isinstance(ret, _GenericAlias) and ret.__origin__ == Generator:
        return ret.__args__[2], wrapper
    else:
        return ret, wrapper


Func = TypeVar("Func", bound=Callable)


class MethodConverter:
    def __init__(self, decorator: Callable, func: Callable):
        self.decorator = decorator
        self.func = func

    def __get__(self, instance, owner):
        return self if instance is None else MethodType(self.func, instance)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __set_name__(self, owner, name):
        self.decorator(self.func, owner)


def handle_method(decorator: Func) -> Func:
    def decorator2(func, param=None, ret=None):
        if param is None and ret is None:
            if list(signature(func).parameters)[0] == "self":
                return MethodConverter(decorator, func)
        return decorator(func, param, ret)

    return cast(Func, decorator2)


Cls = TypeVar("Cls", bound=Type)
Other = TypeVar("Other", bound=Type)


def _get_origin(cls: _GenericAlias) -> Type:  # type: ignore
    return TYPED_ORIGINS.get(cls.__origin__, cls.__origin__)  # type: ignore


# If A[str, T] and A[T, int], what to choose for A[str, int]? None of them.

def substitute_generic_args(cls: Cls, substitution: Mapping) -> Cls:
    if not isinstance(cls, _GenericAlias):
        return cls
    args = tuple(substitution.get(arg, arg) for arg in cls.__args__)
    return _get_origin(cls)[args]


def substitute_type_vars(base: Cls, other: Other) -> Tuple[Cls, Other]:
    if not isinstance(base, _GenericAlias):
        return base, other
    substitution = {
        arg: param
        for arg, param in zip(base.__args__, _get_origin(base).__parameters__)
        if isinstance(arg, TypeVar)  # type: ignore
    }
    if isinstance(other, TypeVar):  # type: ignore
        new_other = substitution.get(other, other)
    else:
        new_other = substitute_generic_args(other, substitution)
    if all(isinstance(arg, TypeVar) for arg in base.__args__):  # type: ignore
        return cast(Tuple[Cls, Other], (_get_origin(base), new_other))
    else:
        return cast(Tuple[Cls, Other],
                    (substitute_generic_args(base, substitution), new_other))


@handle_method
def converter(function: Func, param: Type = None, ret: Type = None) -> Func:
    param, ret = check_converter(function, param, ret)
    param, ret = substitute_type_vars(param, ret)
    _converters[param][ret] = function
    return function


@handle_method
def input_converter(function: Func, param: Type = None, ret: Type = None
                    ) -> Func:
    param, ret = check_converter(function, param, ret)
    ret, function_ = handle_potential_validation(ret, function)
    ret, param = substitute_type_vars(ret, param)
    _input_converters[ret][param] = function_
    return function


@handle_method
def output_converter(function: Func, param: Type = None, ret: Type = None
                     ) -> Func:
    param, ret = check_converter(function, param, ret)
    param, ret = substitute_type_vars(param, ret)
    _output_converters[param] = ret, function
    _converters[param][ret] = function
    return function


def inout_model(model: Type, *, replace_new_type: bool = True
                ) -> Callable[[Cls], Cls]:
    def decorator(cls: Cls) -> Cls:
        input_converter(cls, model, cls)
        out_conv = model
        if replace_new_type and hasattr(model, "__supertype__"):
            out_conv = model.__supertype__
        output_converter(out_conv, cls, model)
        return cls

    return decorator


Arg = TypeVar("Arg")


class InputVisitorMixin(Generic[Arg, Return]):
    def _custom(self, cls: Type, custom: Dict[Type, Converter],
                arg: Arg) -> Return:
        raise NotImplementedError()

    def custom(self, cls: Type, arg: Arg) -> Union[Return, NotCustom]:
        if cls not in _input_converters:
            return NOT_CUSTOM
        return self._custom(cls, _input_converters[cls], arg)


class OutputVisitorMixin(Generic[Arg, Return]):
    def __init__(self, conversions: Mapping[Type, Type]):
        self.converters = {}
        for cls, out in conversions.items():
            cls, out = substitute_type_vars(cls, out)
            if out not in _converters.get(cls, {}):
                logger.warning(f"{cls} cannot be converted to {out}")
            else:
                self.converters[cls] = out, _converters[cls][out]

    def _custom(self, cls: Type, custom: Tuple[Type, Converter],
                arg: Arg) -> Return:
        raise NotImplementedError()

    def custom(self, cls: Type, arg: Arg) -> Union[Return, NotCustom]:
        if cls in self.converters:
            return self._custom(cls, self.converters[cls], arg)
        if cls in _output_converters:
            return self._custom(cls, _output_converters[cls], arg)
        return NOT_CUSTOM


INPUT_METADATA = f"{PREFIX}input"
OUTPUT_METADATA = f"{PREFIX}output"


def field_input_converter(converter: Callable[[Any], Any], param: Type = None,
                          ret: Type = None) -> Metadata:
    try:
        param, ret = check_converter(converter, param, ret)
    except NameError:
        pass
    else:
        ret, converter = handle_potential_validation(ret, converter)
    return DictWithUnion({INPUT_METADATA: (param, converter)})


def field_output_converter(converter: Callable[[Any], Any], param: Type = None,
                           ret: Type = None) -> Metadata:
    try:
        param, ret = check_converter(converter, param, ret)
    except NameError:
        pass
    return DictWithUnion({OUTPUT_METADATA: (ret, converter)})


# Because of NameError when type is recursive
def resolve_fieds_converters(cls: Type):
    assert is_dataclass(cls)
    for field in fields(cls):
        metadata = dict(field.metadata)
        if INPUT_METADATA in metadata:
            param, converter = metadata[INPUT_METADATA]
            param, _ = check_converter(converter, param, ...)  # type: ignore
            param, _ = substitute_type_vars(param, ...)  # type: ignore
            _, converter = handle_potential_validation(..., converter)  # type: ignore
            metadata[INPUT_METADATA] = param, converter
        if OUTPUT_METADATA in metadata:
            ret, converter = metadata[OUTPUT_METADATA]
            _, ret = check_converter(converter, ..., ret)  # type: ignore
            ret, _ = substitute_type_vars(ret, ...)  # type: ignore
            metadata[OUTPUT_METADATA] = ret, converter
        field.metadata = MappingProxyType(metadata)


def converter_from_raw(func: Callable) -> Converter:
    types = get_type_hints(func)
    sig = signature(func)
    annotations: Dict[str, Any] = {}
    fields: Dict[str, Any] = {}
    kwargs = None
    for name, param in sig.parameters.items():
        if param.kind == Parameter.POSITIONAL_ONLY:
            raise TypeError(f"Forbidden positional-only parameter")
        if param.kind == Parameter.VAR_POSITIONAL:
            raise TypeError(f"Forbidden variadic positional parameter")
        if param.kind == Parameter.VAR_KEYWORD:
            fields[name] = field(default_factory=dict,  # type: ignore
                                 metadata=properties())
            type_ = types.get(name, Any)
            annotations[name] = Mapping[str, type_]  # type: ignore
            kwargs = name
            continue
        try:
            default = MISSING
            if param.default != Parameter.empty:
                default = param.default
            fields[name] = default
            annotations[name] = types[name]
        except KeyError:
            raise TypeError("All parameters must be annotated")

    def converter(obj):
        kw = as_dict(obj)
        kw.update(kw.pop(kwargs, ()))
        return func(**kw)

    namespace = {**fields, "__annotations__": annotations}  # type: ignore
    cls: Type = dataclass(type(to_camel_case(func.__name__), (), namespace))
    converter.__annotations__ = {"obj": cls, "return": types.get("return")}
    return converter


def raw_input_converter(func: Func) -> Func:
    input_converter(converter_from_raw(func))
    return func


def raw_field_input_converter(func: Callable) -> Metadata:
    return field_input_converter(raw_input_converter(func))



from dataclasses import dataclass, field
from enum import Enum
from typing import Mapping

from apischema import from_stringified, properties
from apischema.typing import Literal


class Strength(Enum):
    WEAK = "weak"
    STRONG = "strong"


Size = Literal["small", "big"]


@dataclass
class Config:
    strength: Strength = Strength.WEAK
    size: Size = "small"
    level: int = 0
    extra: Mapping[str, bool] = field(default_factory=dict,
                                      metadata=properties())


def test_from_stringified():
    lines = {"level": "42", "opt1": "ok", "opt2": "no"}
    assert from_stringified(lines.items(), Config) == Config(
        strength=Strength.WEAK,
        size="small",
        level=42,
        extra={"opt1": True, "opt2": False}
    )



from __future__ import annotations

from typing import Collection

from apischema import converter, output_converter, to_data
from apischema.typing import TypedDict


class LightSerialization(TypedDict):
    id: int
    owner: str


class FullSerialization(TypedDict):
    id: int
    owner: str
    tags: Collection[str]


class OrmEntity:
    def __init__(self, id: int, owner: str, tags: Collection[str]):
        self.id = id
        self.owner = owner
        self.tags = tags

    @converter
    def to_dict_light(self) -> LightSerialization:
        return {"id": self.id, "owner": self.owner}

    @output_converter
    def to_dict(self) -> FullSerialization:
        return {"id": self.id, "owner": self.owner, "tags": self.tags}


def test_conversion():
    entity1 = OrmEntity(1, "me", ["tag"])
    entity2 = OrmEntity(2, "me", [])
    assert to_data(entity1) == {
        "id":    1,
        "owner": "me",
        "tags":  ["tag"]
    } == to_data(entity1, conversions={OrmEntity: FullSerialization})
    entities = [entity1, entity2]
    assert to_data(entities, conversions={OrmEntity: LightSerialization}) == [
        {
            "id":    1,
            "owner": "me"
        },
        {
            "id":    2,
            "owner": "me"
        }
    ]



from dataclasses import InitVar, dataclass, field, fields

from pytest import raises

from apischema import (fields_set, mark_set_fields, unmark_set_fields,
                       with_fields_set)
from apischema.fields import (FIELDS_SET_ATTR, NoDefault, get_default,
                              has_default, init_fields)


@with_fields_set
@dataclass
class Data:
    without_default: int
    with_default: int = 0
    with_default_factory: int = field(default_factory=lambda: 0)


def test_default():
    without_default, with_default, with_default_factory = fields(Data)
    assert not has_default(without_default)
    assert has_default(with_default)
    assert has_default(with_default_factory)

    with raises(NoDefault):
        get_default(without_default)
    assert get_default(with_default) == 0
    assert get_default(with_default_factory) == 0


def test_fields_set():
    with raises(TypeError):
        fields_set(object())

    assert fields_set(Data(0)) == {"without_default"}
    assert fields_set(Data(without_default=0)) == {"without_default"}
    assert fields_set(Data(0, 1)) == {"without_default", "with_default"}

    data = Data(0)
    data.with_default = 1
    assert fields_set(data) == {"without_default", "with_default"}
    unmark_set_fields(data, "without_default")
    assert fields_set(data) == {"with_default"}
    mark_set_fields(data, "with_default_factory")
    assert fields_set(data) == {"with_default", "with_default_factory"}
    mark_set_fields(data, "with_default", overwrite=True)
    assert fields_set(data) == {"with_default"}
    data.__dict__.pop(FIELDS_SET_ATTR)
    assert fields_set(data) == {"without_default", "with_default",
                                "with_default_factory"}
    unmark_set_fields(data, "without_default")
    assert fields_set(data) == {"with_default", "with_default_factory"}
    data.__dict__.pop(FIELDS_SET_ATTR)
    mark_set_fields(data, "with_default")
    assert fields_set(data) == {"without_default", "with_default",
                                "with_default_factory"}
    mark_set_fields(data, "other")
    assert fields_set(data) == {"without_default", "with_default",
                                "with_default_factory", "other"}

    @with_fields_set(init=False)
    class Data2:
        def __init__(self, a: int):
            self.field = a

    assert fields_set(Data2(0)) == {"field"}


@dataclass
class InitData:
    value: int
    init: InitVar[int]
    no_init: int = field(init=False)

    def __post_init__(self, init: int):
        self.no_init = init


def test_init_fields():
    assert [f.name for f in init_fields(InitData)] == ["value", "init"]
    assert [f.name for f in fields(InitData)] == ["value", "no_init"]



from dataclasses import dataclass, field
from typing import List, Mapping, NewType

from apischema import from_stringified, output_converter

Secret = NewType("Secret", str)


@output_converter
def hide_secret(_: Secret) -> str:
    return "******"


@dataclass
class Config:
    username: str
    password: Secret
    options: Mapping[str, bool] = field(default_factory=dict)
    authorized_domains: List[str] = field(default_factory=list)


def test_config():
    raw_data = """
    username=wyfo
    password=5tr0ngP455w0rd!
    options.verbose=true
    options.execute_order_66=no
    authorized_domains.0=fr
    authorized_domains.1=com
    """
    key_values = [
        tuple(line.split("="))
        for line in map(str.strip, raw_data.split("\n"))
        if line
    ]
    assert from_stringified(key_values, Config) == Config(
        username="wyfo",
        password=Secret("5tr0ngP455w0rd!"),
        options={"verbose": True, "execute_order_66": False},
        authorized_domains=["fr", "com"]
    )



from setuptools import find_packages, setup

with open("README.md") as f:
    README = f.read()

# cannot use Cython because of https://github.com/cython/cython/issues/3537
setup(
    name='apischema',
    url="https://github.com/wyfo/apischema",
    author="Joseph Perez",
    author_email="joperez@hotmail.fr",
    description="Another Python API schema handling and JSON (de)serialization "
                "through typing annotation; light, simple, powerful.",
    long_description=README,
    long_description_content_type="text/markdown",
    version='0.1.2',
    packages=find_packages(include=["apischema*"]),
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    python_requires='>=3.7'
)



import sys
from datetime import date, datetime, time
from ipaddress import (IPv4Address, IPv4Interface, IPv4Network,
                       IPv6Address, IPv6Interface, IPv6Network)
from uuid import UUID

from apischema.conversion import inout_model, input_converter, output_converter
from apischema.schema import schema

# ==================== uuid ====================

inout_model(str)(UUID)
schema(format="uuid")(UUID)

# ==================== datetime ====================

if sys.version_info >= (3, 7):
    input_converter(date.fromisoformat, str, date)
    output_converter(date.isoformat, date, str)
    schema(format="date")(date)

    input_converter(datetime.fromisoformat, str, datetime)
    output_converter(datetime.isoformat, datetime, str)
    schema(format="date-time")(datetime)

    input_converter(time.fromisoformat, str, time)
    input_converter(time.isoformat, time, str)
    schema(format="time")(time)

# ==================== ipaddress ====================

inout_model(str)(IPv4Address)
schema(format="ipv4")(IPv4Address)

inout_model(str)(IPv4Interface)
schema(format="ipv4")(IPv4Interface)

inout_model(str)(IPv4Network)
schema(format="ipv4")(IPv4Network)

inout_model(str)(IPv6Address)
schema(format="ipv6")(IPv6Address)

inout_model(str)(IPv6Interface)
schema(format="ipv6")(IPv6Interface)

inout_model(str)(IPv6Network)
schema(format="ipv6")(IPv6Network)



from dataclasses import Field
from functools import wraps
from inspect import Parameter, isfunction, isgeneratorfunction, signature
from types import MethodType
from typing import (AbstractSet, Any, Callable, Iterable,
                    Iterator,
                    Sequence,
                    TypeVar, overload)

from apischema.types import DictWithUnion, Metadata
from apischema.utils import PREFIX, to_generator
from apischema.validation.dependencies import find_end_dependencies
from apischema.validation.errors import (Error, ValidationError,
                                         build_from_errors, exception)
from apischema.validation.mock import NonTrivialDependency

T = TypeVar("T", contravariant=True)


# use attribute instead of global dict in order to be inherited
VALIDATORS_ATTR = f"{PREFIX}validators"
ValidatorResult = Iterator[Error]


class Validator:
    def __init__(self, func: Callable):
        wraps(func)(self)
        self.func = func
        if not isgeneratorfunction(func):
            self.validate = to_generator(func)
        else:
            self.validate = func

    def __get__(self, instance, owner):
        return self if instance is None else MethodType(self.func, instance)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __set_name__(self, owner, name: str):
        parameters = signature(self.func).parameters
        if not any(param.kind == Parameter.VAR_KEYWORD
                   for param in parameters.values()):
            params = set(parameters) - {"self"}
            wrapped = self.validate

            def wrapper(self: T, **kwargs):
                kwargs2 = {k: v for k, v in kwargs.items() if k in params}
                return wrapped(self, **kwargs2)

            self.validate = wrapper
        self.owner = owner
        self.dependencies = find_end_dependencies(owner, self.func)
        setattr(owner, VALIDATORS_ATTR, (*get_validators(owner), self))

    def can_be_called(self, fields: AbstractSet[str]) -> bool:
        return all(dep in fields for dep in self.dependencies)


def get_validators(obj) -> Sequence[Validator]:
    return getattr(obj, VALIDATORS_ATTR, ())


class Discard(Exception):
    def __init__(self, *fields):
        if not all(isinstance(f, Field) for f in fields):
            raise TypeError("Only fields can be discarded")
        self.fields: AbstractSet[str] = {f.name for f in fields}


def validate(_obj: T, _validators: Sequence[Validator] = None, **kwargs):
    if _validators is None:
        _validators = get_validators(_obj)
    if kwargs is None:
        kwargs = {}
    errors = []
    while True:
        for i, validator in enumerate(_validators):
            try:
                errors.extend(validator.validate(_obj, **kwargs))
            except ValidationError:
                raise
            except Discard as err:
                _validators = [v for v in _validators[i + 1:]
                               if not v.dependencies & err.fields]
                break
            except NonTrivialDependency as exc:
                exc.validator = validator
                raise
            except Exception as err:
                errors.append(exception(err))
        else:
            break
    if errors:
        raise build_from_errors(errors)


V = TypeVar("V", bound=Callable)


@overload
def validator(func: V) -> V:
    ...


@overload
def validator(field: Any, *, discard: bool = True) -> Callable[[V], V]:
    ...


def validator(arg=None, field=None, *, discard=True):
    if isfunction(arg):
        return Validator(arg)
    if arg is None and field is None or arg is not None and field is not None:
        raise ValueError("Bad use of validator")
    field = field or arg
    if not isinstance(field, Field):
        raise TypeError("validator argument must be a field")

    def decorator(func):
        @wraps(func)
        def wrapper(self, **kwargs):
            has_error = False
            for error in func(self, **kwargs):
                has_error = True
                if isinstance(error, str):
                    yield field, error
                else:
                    path, msg = error
                    if isinstance(path, Iterable) and not isinstance(path, str):
                        yield (field, *path), msg
                    else:
                        yield (field, path), msg
            if has_error and discard:
                raise Discard(field)

        return validator(wrapper)

    return decorator


VALIDATORS_METADATA = f"{PREFIX}validators"


def field_validator(*validators: Callable) -> Metadata:
    return DictWithUnion({VALIDATORS_METADATA: validators})



import ast
from inspect import getsourcelines, isfunction
from itertools import takewhile
from typing import Callable, ClassVar, Dict, Mapping, Set, Type, get_type_hints


def getsource(func: Callable) -> str:
    lines, _ = getsourcelines(func)
    indentation = sum(1 for _ in takewhile(str.isspace, lines[0]))
    return "".join(line[indentation:] for line in lines)


Dependency = str
Dependencies = Set[Dependency]


class DependencyFinder(ast.NodeVisitor):
    def __init__(self):
        self.dependencies: Set[str] = set()

    def visit_Attribute(self, node):
        self.generic_visit(node)
        if isinstance(node.value, ast.Name) and node.value.id == "self":
            self.dependencies.add(node.attr)


def find_dependencies(method: Callable) -> Dependencies:
    finder = DependencyFinder()
    try:
        finder.visit(ast.parse(getsource(method)))
    except TypeError:
        return set()
    return finder.dependencies


cache: Dict[Callable, Dependencies] = {}


def is_class_var(dep: str, cls: Type, type_hints: Mapping[str, Type]) -> bool:
    # Cannot use is_dataclass because dependencies of dataclass are evaluated
    # before the dataclass decorator
    if not hasattr(cls, dep):
        return False
    if dep in type_hints:
        return getattr(type_hints[dep], "__origin__",
                       type_hints[dep]) is ClassVar
    return True


def find_end_dependencies(cls: type, method: Callable, rec_gard=()
                          ) -> Dependencies:
    if method not in cache:
        type_hints = get_type_hints(cls)
        dependencies = find_dependencies(method)
        class_vars = {dep for dep in dependencies
                      if is_class_var(dep, cls, type_hints)}
        dependencies -= class_vars
        for field in class_vars:
            attr = getattr(cls, field)
            if isinstance(attr, property):
                attr = attr.fget
            if isfunction(attr):
                if attr in rec_gard:
                    continue
                rec_deps = find_end_dependencies(cls, attr, (*rec_gard, attr))
                dependencies.update(rec_deps)
        cache[method] = dependencies
    return cache[method]
